# Copyright (C) 2004-2005 Marcin Feder, Michal J. Pietal & GeneSilico :: {marcin,michalp}@genesilico.pl
# uses:    AATranslator module/class, Python 2.4 and BioPython 1.3

__doc__=""" 
This module contains BrutePDBParser - the Bio.PDB.PDBParser specialized and PDB error-resistant subclass implementation.

"""

import sys
from string import split, ascii_letters, atoi, join, digits, strip#todo: exchange for ascii_upper
from Numeric import array, Float0
import types

from Bio.PDB.PDBParser import PDBParser

from genesilico.structure.AATranslator import AATranslator
from genesilico.structure.BruteStructureBuilder import BruteStructureBuilder
from genesilico.structure.BruteStructureBuilder import _b_parse_pdb_header_list as _parse_pdb_header_list
try:
    from genesilico.structure.SMCRA_extended import get_ligands_list, get_ions_list
except ImportError:
    pass

from Bio.PDB.PDBExceptions import PDBConstructionException, PDBException #strings
from genesilico.structure.BPDBExceptions import BPDBConstructionException, BPDBProteinException, BPDBException, BPDBxNAException #classes

################################################################################################################

class BrutePDBParser(PDBParser):
    """
    Parse a PDB file and return a Structure object.
    
    It allows doing it in three different ways:
    -classic (PDBParser)
    -through atomic data list (get_coords)
    -ignoring non- protein chains (get_protein structure).
    
    It also has ability (via two flags added) to fix two different PDB file corruption types:
    -lack of chain id character
    -(locally) misnumbered residue id
    
    Furthermore, 'end tag' variable is introduced to handle various non-converted (to 'ENDMDL') ensembles.
    
    It can also parse giant files (ensembles) while parsing stops after first model end is reached.
    
    Additional changes include residue.secondary_pdb entries in returned structure and BrutePDBParser.seqres list,
    both derived from pdb file header.  Though, this is temporal in terms of API, not functionality.
    
    """
    
    def __init__(self, PERMISSIVE=1,get_header=0,structure_builder=None,FIXCHAINID=0,FIXRESID=0,SINGLEMODEL=0,UNMODIFIED_RESNAME=0,CLEAN_ICODE=0):
        """
        The PDB parser call a number of standard methods in an aggregated
        StructureBuilder object. Normally this object is instanciated by the
        PDBParser object itself, but if the user provides his own StructureBuilder
        object, the latter is used instead.

        Arguments:
        o PERMISSIVE - int;
            0 exceptions in constructing the SMCRA data structure are fatal. 
            1 (DEFAULT), the exceptions are caught and stored in a log variable, but some residues or atoms will be missing. 
            2 - same as above but all printed to stdout
            THESE EXCEPTIONS ARE DUE TO PROBLEMS IN THE PDB FILE!.
        Option 2 causes the exceptional info gets printed out.
        o structure_builder - an optional user implemented StructureBuilder class. 
        o FIXCHAINID - int, fills in the blank chain id after a chain letter while multi-chain model in a PDB file
        o FIXRESID - int, corrects the mistyped residue id (number)
        o SINGLEMODEL - int, imports first model from the file only
        o UNMODIFIED_RESNAME - int - changes name of modified residue back to original	
        o CLEAN_ICODE - int - while non-empty insertion code encountered, increases residue's resseq instead
        The two last flags can be used together but it is recommented to try them separately first while parsing 
        currupted pdb file.
        """
        if structure_builder!=None:
            self.structure_builder=structure_builder
            self.parsing_log= None
        else:
            self.structure_builder=BruteStructureBuilder()
            self.parsing_log= True
        self.header=None
        self.trailer=None
        self.line_counter=0
        
        self.PERMISSIVE=PERMISSIVE
        self.FIXCHAINID=FIXCHAINID
        self.FIXRESID=FIXRESID
        self.SINGLEMODEL=SINGLEMODEL
        self.UNMODIFIED_RESNAME = UNMODIFIED_RESNAME
        self.CLEAN_ICODE = CLEAN_ICODE
        
        self.translator=AATranslator(unknown3letter='UNK')
        
    def _parse(self, header_coords_trailer, endtag, id):
        "Parse the PDB file."
        # Extract the header; return the rest of the file
        self.header, coords_trailer, secondary =self._get_header(header_coords_trailer)
        # Parse the atomic data; return the PDB file trailer
        if id == None:
            id=self.header['pdbid']
            if id == None and self.PERMISSIVE == 0:
                #no id or invalid file
                raise BPDBException, "No HEADER record.  Structure id should be given as first argument"
        # Make a StructureBuilder instance (pass id of structure as parameter)
        self.structure_builder.init_structure(id)
        self.trailer=self._parse_coordinates(coords_trailer, endtag)
        self.secondary = self._insert_secondary(secondary)
        
    def _get_header(self, header_coords_trailer):
        """
        Get the header of the PDB file, return the rest, also reading the secondary structure elements 
        (HELIX/SHEET/TURN) into the temporary data structure.
      
        First two positions of each secondary-tripple determine the structure range (as first char is a chain_id, 
        the remaining four are residue_ids), third tells about secondary structure kind (single character).
        """
        secondary=[]
        structure_builder=self.structure_builder
        i=0
        for i in range(0, len(header_coords_trailer)):
            structure_builder.set_line_counter(i+1)
            line=header_coords_trailer[i]
            record_type=line[0:6]
            if(record_type=='ATOM  ' or record_type=='HETATM' or record_type=='MODEL '):
                break
            elif record_type == 'HELIX ':
                secondary.append((line[19]+line[21:25],line[31]+line[33:37],'H'))
            elif record_type == 'SHEET ':
                secondary.append((line[21]+line[22:26],line[32]+line[33:37],'S'))
            elif record_type == 'TURN  ':
                secondary.append((line[19]+line[20:24],line[30]+line[31:35],'T'))
        
        header=header_coords_trailer[0:i]
        self.seqres=self._parse_seqres(header)
        # Return the rest of the coords+trailer for further processing
        self.line_counter=i
        coords_trailer=header_coords_trailer[i:]
        
        #if self.parsing_log != None:
        header_dict, header_log=_parse_pdb_header_list(header,self.PERMISSIVE)
        if self.parsing_log != None:
            #silent parsing
            self.structure_builder.expand_log(header_log)
        else:
            if header_log != '':
                print header_log
        #else:
        #    header_dict=parse_pdb_header_list(header)
        
        return header_dict, coords_trailer, secondary
    
    def _parse_coordinates(self, coords_trailer, endtag):
        """
        Parse the atomic data in the PDB file.
        
        FIXRESID flag corrects the small resseq order perturbations, e.g. 1 2 3 2 5 6...
        FIXCHAINID inserts character chain id when blank field encountered (like 'A151' 'A152' '   1' '   2')
        """
        aa_codes =  self.translator.get_codes_extended()
        
        local_line_counter=0
        structure_builder=self.structure_builder
        current_model_id=0
        current_chain_id=None
        current_segid=None
        current_residue_id=None
        current_resname=None
        structure_builder.init_model(current_model_id)
        
        for i in range(0, len(coords_trailer)):
            line=coords_trailer[i]
            
            length=len(endtag)+1
            record_max=line[0:length]
            record_type=line[0:6]
            
            global_line_counter=self.line_counter+local_line_counter+1
            structure_builder.set_line_counter(global_line_counter)
            if(record_type=='ATOM  ' or record_type=='HETATM'):
                fullname=line[12:16]
                # get rid of whitespace in atom names
                split_list=split(fullname)
                if len(split_list)!=1:
                    # atom name has internal spaces, e.g. " N B ", so
                    # we do not strip spaces
                    name=fullname
                else:
                    # atom name is like " CA ", so we can strip spaces
                    name=split_list[0]
                altloc=line[16:17]
                resname=line[17:20]
                if self.UNMODIFIED_RESNAME == True:
                    un_resname = self.translator.get33original(resname) 
                    if un_resname != self.translator.unknown3:
                        resname=un_resname
                chainid=line[21:22]
                try:
                    serial_number=int(line[6:11])
                except:
                    serial_number=0
                resseq=int(split(line[22:26])[0])   # sequence identifier   
                
                icode=line[26:27]           # insertion code
                
                #FIX* mandatory code
                
                if self.CLEAN_ICODE==1:
                    if icode != ' ':
                        if old_icode != icode:
                            add_resseq = add_resseq + (old_resseq - resseq - add_resseq)+1
                        old_icode = icode
                        icode = ' '
                    else:
                        old_icode = icode
                
                if current_chain_id == None:
                    #first chain first residue
                    chain_alphabet=list(ascii_letters[26:])
                    altloc_alphabet=list(ascii_letters[26:])
                    
                    base_resseq = resseq
                    add_resseq = 0
                    old_chainid = chainid
                    
                    if self.FIXCHAINID == 1:
                        if chainid not in chain_alphabet:
                            old_chainid = chainid
                            chainid = chain_alphabet[0]
                        else:
                            #chainid --> chain_alphabet[0]
                            chain_alphabet = [chainid]+chain_alphabet[:chain_alphabet.index(chainid)]+chain_alphabet[chain_alphabet.index(chainid)+1:]
                else:
                    # FIXRESID mode corrects misnumbered residue numbers
                    if self.FIXRESID == 1:
                        if chainid != old_chainid and resseq+add_resseq > old_resseq:
                            #new chain
                            resseq = base_resseq
                            add_resseq = 0
                        elif chainid == old_chainid and resseq+add_resseq < old_resseq: 
                            #misnumbered - ++
                            #    add_resseq = add_resseq + 1
                            add_resseq = add_resseq + (old_resseq - resseq - add_resseq)+1
                        elif chainid == old_chainid and (resseq+add_resseq == old_resseq and resname != old_resname and altloc == ' ' and icode == ' '):
                            #misnumbered - ++
                            add_resseq = add_resseq + 1
                    
                    old_chainid = chainid
                    
                    # FIXCHAINID fills in the blank chainid field
                    if self.FIXCHAINID == 1:
                        if chainid == old_chainid and resseq+add_resseq < old_resseq:
                            #new chain encountered
                            del chain_alphabet[0]
                            chainid = chain_alphabet[0]
                            add_resseq = 0
                        elif self.FIXRESID == 0 and chainid == old_chainid and (resseq+add_resseq == old_resseq and resname != old_resname and altloc == ' ' and icode == ' '):
                            #new chain encountered
                            del chain_alphabet[0]
                            chainid = chain_alphabet[0]
                            add_resseq = 0
                        elif chainid == old_chainid and chainid != current_chain_id:
                            #rename a chain within
                            chainid = chain_alphabet[0]
                        elif chainid != old_chainid:
                            #regular new chain - rename if non-letter
                            del chain_alphabet[0]
                            if chainid not in chain_alphabet:
                                chainid = chain_alphabet[0]
                            else:
                                #chainid --> chain_alphabet[0]
                                chain_alphabet = [chainid]+chain_alphabet[:chain_alphabet.index(chainid)]+chain_alphabet[chain_alphabet.index(chainid)+1:]
                            add_resseq = 0
                
                resseq = resseq + add_resseq
                old_resname = resname
                if icode == ' ':
                    old_resseq = resseq
                
                if record_type=='HETATM':       # hetero atom flag
                    if resname=="HOH" or resname=="WAT":
                        hetero_flag="W"
                    elif self.UNMODIFIED_RESNAME == True and resname in aa_codes:
                        #here hetatm group is treated as a modified residue
                        hetero_flag=" "
                    else:
                        hetero_flag="H"
                    
                else:
                    hetero_flag=" "
                
                residue_id=(hetero_flag, resseq, icode)
                
                # atomic coordinates
                x=float(line[30:38]) 
                y=float(line[38:46]) 
                z=float(line[46:54])
                coord=array((x, y, z), Float0)
                # occupancy & B factor
                try:
                    occupancy=float(line[54:60])
                except ValueError:
                    occupancy=0.00
                try:    
                    bfactor=float(line[60:66])
                except ValueError:
                    bfactor=0.00
		#try:
		#    element=str(line[76:78])
		#except ValueError:
		#    element=' '+name.strip()[0]
		element=name.strip()[0]
                segid=line[72:76]
                if current_segid!=segid:
                    current_segid=segid
                    structure_builder.init_seg(current_segid)
                if current_chain_id!=chainid:
                    current_chain_id=chainid
                    structure_builder.init_chain(current_chain_id)
                    current_residue_id=residue_id
                    current_resname=resname
                    try:
                        structure_builder.init_residue(resname, hetero_flag, resseq, icode)
                    except (BPDBConstructionException,PDBConstructionException), message:
                        self._handle_PDB_exception(message, global_line_counter)
                    except (BPDBException,PDBException), message:
                        self._handle_PDB_exception(message, global_line_counter)
                elif current_residue_id!=residue_id or current_resname!=resname:
                    current_residue_id=residue_id
                    current_resname=resname
                    try:
                        structure_builder.init_residue(resname, hetero_flag, resseq, icode)
                    except (BPDBConstructionException,PDBConstructionException), message:
                        self._handle_PDB_exception(message, global_line_counter)
                    except (BPDBException,PDBException), message:
                        self._handle_PDB_exception(message, global_line_counter)
                # init atom
                try:
                    #new bio version - without serial_number is tried first
                    structure_builder.init_atom(name, coord, bfactor, occupancy, altloc, fullname, serial_number, element)
                except (BPDBConstructionException,PDBConstructionException), message:
                    self._handle_PDB_exception(message, global_line_counter)
                except (BPDBException,PDBException), message:
                        self._handle_PDB_exception(message, global_line_counter)
                
            elif(record_type=='ANISOU'):
                anisou=map(float, (line[28:35], line[35:42], line[43:49], line[49:56], line[56:63], line[63:70]))
                # U's are scaled by 10^4 
                anisou_array=(array(anisou, Float0)/10000.0).astype(Float0)
                structure_builder.set_anisou(anisou_array)
            elif (record_max in [endtag+' ',endtag+'\n'] or record_max[:-1] == line):
                if self.SINGLEMODEL==0:
                    if not (len(coords_trailer)>i+1 and coords_trailer[i+1][0:6] in ['END   ','CONECT']):
                        current_model_id=current_model_id+1
                        structure_builder.init_model(current_model_id)
                        current_chain_id=None
                        current_residue_id=None
                else:
                    #end of model == end of structure
                    self.line_counter=self.line_counter+local_line_counter
                    return coords_trailer[local_line_counter:]
            elif(record_type=='END   ' or record_type=='CONECT'):
                # End of atomic data, return the trailer
                self.line_counter=self.line_counter+local_line_counter
                return coords_trailer[local_line_counter:]
            elif(record_type=='SIGUIJ'):
                # standard deviation of anisotropic B factor
                siguij=map(float, (line[28:35], line[35:42], line[42:49], line[49:56], line[56:63], line[63:70]))
                # U sigma's are scaled by 10^4
                siguij_array=(array(siguij, Float0)/10000.0).astype(Float0)   
                structure_builder.set_siguij(siguij_array)
            elif(record_type=='SIGATM'):
                # standard deviation of atomic positions
                sigatm=map(float, (line[30:38], line[38:45], line[46:54], line[54:60], line[60:66]))
                sigatm_array=array(sigatm, Float0)
                structure_builder.set_sigatm(sigatm_array)
            elif(record_type == 'TER   '):
                #FIXCHAINID and FIXRESID mandatory code
                old_resseq = resseq
                old_chainid = chainid
                old_resname = resname
            
            local_line_counter=local_line_counter+1
            
        # EOF (does not end in END or CONECT)
        self.line_counter=self.line_counter+local_line_counter
        return []
    
    def _parse_seqres(self, header):
        """
        The method parses seqres pdb header entries.  This approach is deprecated due to PDBParser developement and will soon be 
        changed or moved elsewhere.
        
        Tests, indeed, prove this works in general but level of complication and contradictive noisy pdb examples suggest it should 
        be re-engineered, improved and massively tested, in order to become a valuable part of the parser.
        
        The seqres range, stored in self.dbref variable for example, needs to be double- checked against structure-derived sequence's
        scope and length while current approach requires doing it outside the parser.
        """
        self.dbref={}
        seqres={}
        
        old_chainid=None
        for i in range(len(header)):
            line=header[i]
            record_type=line[0:6]
            #list of amino acid full- sequence chains
            if record_type == 'DBREF ':
                try:
                    chain_id = line[12]
                    from_begin = int(line[55:60])
                    from_end = int(line[62:67])
                    to_begin = int(line[14:18])
                    to_end = int(line[20:24])
                    if self.dbref.has_key(chain_id):
                       self.dbref[chain_id].append((from_begin,from_end,to_begin,to_end))
                    else:
                        self.dbref[chain_id]=[(from_begin,from_end,to_begin,to_end)]
                except ValueError:
                    print "Value has wrong type, when parsing DBREF"
            elif record_type == 'SEQRES':
                chain_id=line[11]
                #length=line[13:17]
                row=[line[19:22],line[23:26],line[27:30],line[31:34],line[35:38],line[39:42],line[43:46],line[47:50],line[51:54],line[55:58],line[59:62],line[63:66],line[67:70]]
                if chain_id==old_chainid:# and length==old_length:
                    seqres[chain_id]=seqres[chain_id]+row
                else:
                    seqres[chain_id]=row
                old_chainid=chain_id
                #old_length=length
        
        for c in seqres.keys():
            seqres[c]=self.translator.get1letter_seq(join(split(join(seqres[c])),''))
            
        return seqres

    def _insert_secondary(self, secondary_list):
        """
        The method adds the residue.secondary_pdb attribute (H/S/T) in the first model.
        
        The '-' char means either no structure determined or (as a whole) blank pdb secondary entries.
        """
        for chain in self.structure_builder.structure.child_list[0].get_iterator():
            chain_id = chain.get_id()
            for residue in chain.get_iterator():
                residue_id=residue.get_id()[1]
                #residue.secondary_pdb = '-'
                residue.xtra['SS_PDB'] = '-'
                for ss in secondary_list:
                    chain_start = ss[0][0]
                    chain_end = ss[1][0]
                    residue_start =int(ss[0][1:])
                    residue_end =int(ss[1][1:])
                    secondary = ss[2]
                    
                    if chain_start == chain_end:
                        if (chain_id == chain_start and residue_id >= residue_start and residue_id <= residue_end):
                            #residue.secondary_pdb = secondary
                            residue.xtra['SS_PDB'] = secondary
                            break
                    elif (chain_id == chain_start and residue_id >= residue_start) or (chain_id == chain_end and residue_id <= residue_end):
                        #residue.secondary_pdb = secondary
                        residue.xtra['SS_PDB'] = secondary
                        break
        
        self.structure_builder.structure.xtra['has_ss_pdb'] = (len(secondary_list) > 0)
        
        return secondary_list
    
    def get_coords(self, id, coords, endtag='ENDMDL'):
        """
        Return the structure.

        Arguments:
        o id - string, the id that will be used for the structure
        o coords - list of lines of the PDB file
        """
        
        self.header=None
        self.trailer=None
        self._parse(coords,endtag,id)
        # Return the Structure instance
        return self.structure_builder.get_structure()
    
    def get_protein_coords(self, id, coords, endtag='ENDMDL'):
        """
        Return the structure.

        Arguments:
        o id - string, the id that will be used for the structure
        o coords - list of lines of the PDB file
        """
        
        self.header=None
        self.trailer=None
        self._parse(coords, endtag, id)
        
        # Return the protein Structure instance
        base_structure = self.structure_builder.get_structure()
        
        AA_codes_extended = self.translator.get_codes_extended()
        del AA_codes_extended[AA_codes_extended.index('UNK')]
        
        empty = None
        for model in base_structure.get_iterator():
            #empty = None
            for chain in model.get_iterator():
                empty = True
                proper = False
                for residue in chain.get_iterator():
                    if residue.get_id()[0]==' ' and residue.resname in AA_codes_extended:
                        proper = True
                        break
                if not proper:
                    model.detach_child(chain.get_id())
            #empty=True
            for chain in model.get_iterator():
                empty=False
                break
            if empty != False:
                base_structure.detach_child(model.get_id())
        for model in base_structure.get_iterator():
            empty = False
            break
        if empty == True:
            raise BPDBProteinException, self.structure_builder.structure.get_id()
        elif empty == None:
            raise BPDBConstructionException, "No structure"
        
        return base_structure
    
    def get_xna_coords(self, id, coords, endtag='ENDMDL', ligands=False):
        """
        Return the structure.

        Arguments:
        o id - string, the id that will be used for the structure
        o coords - list of lines of the PDB file
        """
        
        self.header=None
        self.trailer=None
        self._parse(coords, endtag, id)
        
        # Return the protein Structure instance
        base_structure = self.structure_builder.get_structure()
        
        AA_codes_extended = self.translator.get_codes_extended()
        del AA_codes_extended[AA_codes_extended.index('UNK')]
        
        ligands_list = get_ligands_list() + get_ions_list()
        wat = ['HOH','WAT']
        
        forbidden = AA_codes_extended + wat
        if not ligands:
            forbidden = forbidden + ligands_list
        
        empty = None
        for model in base_structure.get_iterator():
            for chain in model.get_iterator():
                proper = False
                for residue in chain.get_iterator():
                    if (strip(residue.resname)[-3:] not in forbidden):
                        proper = True
                        break
                if not proper:
                    model.detach_child(chain.get_id())
            empty=True
            for chain in model.get_iterator():
                empty=False
                break
            if empty != False:
                base_structure.detach_child(model.get_id())
        for model in base_structure.get_iterator():
            empty = False
            break
        if empty == True:
            raise BPDBProteinException, self.structure_builder.structure.get_id()
        elif empty == None:
            raise BPDBxNAException, "No structure"
        
        return base_structure
    
    def get_structure(self, id, file=None, endtag='ENDMDL'):
        """Return the structure.

        Arguments:
        o id - string, the id that will be used for the structure
        o file - name of the PDB file OR an open file
        """
        
        if file==None:
            #pdb id as a structure id
            file=id
            id=None
        self.header=None
        self.trailer=None
        if not isinstance(file, types.FileType):
            file=open(file)
        self._parse(file.readlines(), endtag, id)
        file.close()
        self.structure_builder.set_header(self.header)
        # Return the Structure instance
        return self.structure_builder.get_structure()
    
    def get_protein_structure(self, id, file=None, endtag='ENDMDL'):
        """
        Return the structure having polipeptide chains only, thus excluding the rest.
        
        Polipeptyde chain is determined by existence of at least single regular or modified amino acid (code).
        
        Arguments:
        o id - string, the id that will be used for the structure
        o file - name of the PDB file OR an open file
        """
        
        AA_codes_extended = self.translator.get_codes_extended()
        del AA_codes_extended[AA_codes_extended.index('UNK')]
        
        base_structure = self.get_structure(id, file, endtag)
        for model in base_structure.get_iterator():
            for chain in model.get_iterator():
                proper = False
                for residue in chain.get_iterator():
                    if residue.get_id()[0]==' ' and residue.resname in AA_codes_extended:
                        proper = True
                        break
                if not proper:
                    model.detach_child(chain.get_id())
            empty=True
            for chain in model.get_iterator():
                empty=False
                break
            if empty == True:
                base_structure.detach_child(model.get_id())
        empty = True
        for model in base_structure.get_iterator():
            empty = False
            break
        if empty == True:
            #raise PDBConstructionException, "No polipeptyde chains found"
            raise BPDBProteinException, self.structure_builder.structure.get_id()
        return base_structure
    
    def get_xna_structure(self, id, file=None, endtag='ENDMDL'):
        """
        Return the structure having polipeptide chains only, thus excluding the rest.
        
        Polipeptyde chain is determined by existence of at least single regular or modified amino acid (code).
        
        Arguments:
        o id - string, the id that will be used for the structure
        o file - name of the PDB file OR an open file
        """
        
        AA_codes_extended = self.translator.get_codes_extended()
        del AA_codes_extended[AA_codes_extended.index('UNK')]
        
        base_structure = self.get_structure(id, file, endtag)
        for model in base_structure.get_iterator():
            for chain in model.get_iterator():
                proper = False
                for residue in chain.get_iterator():
                    if residue.get_id()[0] in [' ','H'] and residue.resname not in AA_codes_extended:
                        proper = True
                        break
                if not proper:
                    model.detach_child(chain.get_id())
            empty=True
            for chain in model.get_iterator():
                empty=False
                break
            if empty == True:
                base_structure.detach_child(model.get_id())
        empty = True
        for model in base_structure.get_iterator():
            empty = False
            break
        if empty == True:
            #raise PDBConstructionException, "No polipeptyde chains found"
            raise BPDBxNAException, self.structure_builder.structure.get_id()
        return base_structure
    
    def get_parsing_log(self):
        if self.parsing_log == True:
            return self.structure_builder.log
        else:
            return None
    
    def _handle_PDB_exception(self, message, line_counter):
        """
        This method catches an exception that occurs in the StructureBuilder
        object (if PERMISSIVE==1), or raises it again, this time adding the 
        PDB line number to the error message.
        """
        message="%s at line %i." % (message, line_counter)
        if self.PERMISSIVE > 0:
            # just print a warning - some residues/atoms will be missing
            self.structure_builder.log.append("PDBConstructionException: %s" % message)
            self.structure_builder.log.append("Exception ignored.\nSome atoms or residues will be missing in the data structure.")
            if self.PERMISSIVE > 1:
                #stdout
                print "PDBConstructionException: %s" % message
                print "Exception ignored.\nSome atoms or residues will be missing in the data structure."
        else:
            # exceptions are fatal - raise again with new message (including line nr)
            raise BPDBConstructionException, message
        
################################################################################################################

if __name__=="__main__":

    import sys

    if len(sys.argv) == 3:
        PERMISSIVE=int(sys.argv[2])
    else:
        PERMISSIVE=1
    p=BrutePDBParser(PERMISSIVE=PERMISSIVE,FIXCHAINID=0,FIXRESID=0,SINGLEMODEL=0)
    s=p.get_structure("scr", sys.argv[1])
    #for alert in p.get_parsing_log():
    #    print '>',alert

    for m in s.get_iterator():
        p=m.get_parent()
        assert(p is s)
        for c in m.get_iterator():
            p=c.get_parent()
            assert(p is m)
            for r in c.get_iterator():
                p=r.get_parent()
                assert(p is c)
                for a in r.get_iterator():
                    p=a.get_parent()
                    if not p is r:
                        print p, r
