__doc__="GeneSilico answer to Bio.PDB SMCRA classes API"

from types import IntType, StringType, FloatType
from string import split, join, ascii_uppercase, strip
from copy import deepcopy, copy

from Numeric import array

from Bio.PDB.PDBIO import PDBIO as pdbio
from Bio.PDB.PDBIO import Select as select_io
from Bio.Fasta import FastaAlign
from Bio.Alphabet import Gapped
from Bio.Alphabet.IUPAC import IUPACProtein, ExtendedIUPACProtein

# SMCRA hierarchy
from Bio.PDB.Entity import Entity #as entity
from Bio.PDB.Entity import DisorderedEntityWrapper

from Bio.PDB.Structure import Structure as structure
from Bio.PDB.Model import Model as model
from Bio.PDB.Chain import Chain as chain
from Bio.PDB.Residue import Residue as residue
from Bio.PDB.Residue import DisorderedResidue
from Bio.PDB.Atom import Atom as atom
from Bio.PDB.Atom import DisorderedAtom

from Bio.PDB.Polypeptide import CaPPBuilder, PPBuilder, _PPBuilder, Polypeptide
from Bio.PDB.PDBExceptions import PDBException, PDBConstructionException

from BPDBExceptions import BPDBException, BPDBConstructionException
from AATranslator import AATranslator

##########################################################################

def get_ligands_list():

        ligands_list = ['CPT', 'NH2', 'SO4', 'SF4', 'PO4', 'GCK', 'POP', 'PAR', 'DM8', 'DM9', 'BDA', 'DM2', 'PA2',
         'DMY', 'GET', 'MRC', 'Z', 'NGM', 'DM1',  'BCL', 'DPR', 'NAG', 'CNC', 'ABU', 'GLC',
         'MSE', 'DVA', 'NCY', 'FLC', 'NGA', 'ATP', 'DSN', '42B', 'BAL', 'GUN', 'SUC',
         'ACT', 'BTN', 'CH2', 'MOE', 'GOL', 'PG4', 'HPD', 'ARI', 'OTR', 'TER', 'IMT', 'IA', 'HT1',
         'OGG', 'PEG', '7AD', 'EDO', 'GPN', 'DA7', 'HP1', 'DIO', '3DR', 'B9A', 'HP2', '9AD', 'DIM', 'RO2', 
        'C66', 'DAX', 'DM6', 'ERI', 'DIB', 'BFA', 'TNT', 'PXZ', 'AAB', 'OTS', 'A2M', 'BPF', 'IB', 'LAE', 'HT2',
         'CMD', 'PYN', 'DR1', 'SPM', '1GL', 'SAR', 'P24', 'BME', 'LDA', 'C31', 'SPE', 'PTR', 'HT', 'BDG', 'ACY',
         'U31', 'DRC', 'D34', 'T23', 'EPE', 'EOH', 'DRZ', 'D24', 'CPN', 'RHM', 'IIP', 'CDR', 'D18', 'IPA', 
        '104', 'MDN', '8AD', 'PNT', 'E96', 'SPD', 'EDA', 'MPD', 'TRS', 'D35', 'ASU', 'PYB', 'ILT', 'D19', 'PED',
         'TTG', 'QUI', 'MVA', 'YRR', 'TBZ', 'E97', 'SPK', 'IPY', 'CGQ', 'PCA', 'BBZ', 'BDC', 'NOD', 'DA5', 
        'MES', '44D', 'TTC', '232', 'DDX', 'T66', 'A47', 'PET', '9TA', 'DAP', 'ORP', 'IU', 'BRN', 'UCL', 'G49',
         'G36', 'PDI', 'NRI', 'DTV', 'US1', 'NT', 'CPH']

        return ligands_list

def get_ions_list():
    
        ions_list = ['MG','K','NA','NI','CL','CR','CA','SR','BA','V','W','MN','RU','OS','EU','CO','RH','IR','PT','TB',
        'ZN','CD','HG','TL','PB','LU','BR','YB','NCO','3CO','AC']
        
        return ions_list
        
##########################################################################
### Bio bugfix or extended classes
##########################################################################

class Structure(structure):
    
    """This class extends Bio.PDB.Structure SMCRA object, providing additional methods implementation."""


    def __eq__(self, ref2, ref=None):
        """This method compares two SMCRA structures."""
        
        if ref == None:
            ref = self
        
        if not (isinstance(ref2,Entity) or isinstance(ref2,DisorderedEntityWrapper) or isinstance(ref2,Atom) or isinstance(ref2,DisorderedAtom)):
            #type error
            return None
        if not (isinstance(ref,Entity) or isinstance(ref,DisorderedEntityWrapper) or isinstance(ref,Atom) or isinstance(ref,DisorderedAtom)):
            #type error
            return None
        
        #level
        if ref.get_level() != ref2.get_level():
            return False
        
        #different structure names allowed
        if ref.get_level() != 'S':
            #name and id
            if ref.get_id() != ref2.get_id():
                return False
            if ref.get_level() == 'R':
                if ref.get_resname() != ref2.get_resname():
                    return False
            elif ref.get_level() == 'A':
                if ref.get_name() != ref2.get_name():
                    return False
        
        if ref.get_level() == 'A':
            #atom
            profile={}
            for a in [ref, ref2]:
                profile[a]=[a.get_altloc(),a.get_anisou(),a.get_bfactor(),a.get_occupancy(),a.get_sigatm(),a.get_siguij(),\
                a.is_disordered()]+a.get_vector().get_array().tolist()
            for k in range(len(profile[ref])):
                if profile[ref][k] != profile[ref2][k]:
                    return False
            return True
        else:
            #number of children
            if (len(ref.child_list) != len(ref2.child_list)):
                return False
            
            for i in range(len(ref.child_list)):
                if not self.__eq__(ref.child_list[i],ref2.child_list[i]):
                    return False
        return True
        
    
    def _get_chain_if_available(self, chain_id):
        """Get chain by id if exists or get the chain if unique."""
        
        mdl=self.get_prime_model()
        
        chain_no = len(mdl.child_list)
        #any id if single chain
        if chain_no == 1:
            return mdl.child_list[0]
        
        try:
            cha = mdl[chain_id]
        except KeyError, message:
            raise
        
        return cha 
        
    def _map_dict(self, mixed_dict, chain_id):
        """A substitute_* auxilliary method."""
        regular_resid =  {}
        
        for res_key in mixed_dict.keys():
            if type(res_key) == IntType:
                #number
                resseq = res_key
                new_key = (' ',resseq,' ')
            elif type(res_key) == StringType:
                #in string representation
                try:
                    resseq = int(res_key)
                except ValueError:
                    raise KeyError, res_key
                new_key = (' ',resseq,' ')
            else:
                #resseq
                new_key = res_key
            regular_resid[new_key] = float(mixed_dict[res_key])
        
        cha = self._get_chain_if_available(chain_id)
        
        return regular_resid, cha
    
    def add(self, m):
        """Adds model to structure.  
        
        If an empty structure, one can add a chain directly.  A new model is created in such case.
        """
        if isinstance(m,Chain) and len(self.get_list() == 0):
            mdl = Model(0)
            mdl.add(m)
            m = mdl
        try:
            structure.add(self,m)
        except PDBConstructionException, message:
            raise BPDBConstructionException, message
        
        return
    
    def get_prime_model(self):
        """Returns the first model"""
        return self.child_list[0]
    
    def get_parent(self):
        """Returns the parent object
        
        A bugifx to biopython 1.41
        """
        #if self.parent is None and self.get_level() != 'S':
        #    raise BPDBException, 'No parent'
        return self.parent
    
    def get_res_id_list(self, chain_id=None, id_entry=1, SEQRES=0):
        """Get resseq (residue id) list
        
        chain_id might be fixed. SEQRES flag is to inclide missing (no coords) residua.
        """
        cha = self._get_chain_if_available(chain_id)
        
        id_list = []
        for res in cha.get_iterator(SEQRES):
            id_list.append(res.get_id()[id_entry])
        
        return  id_list
    
    def get_sequence(self, chain_id = None,ter_tag='',ORIGINAL=0, SEQRES=0, UNKNOWN_ALLOWED=1):
        """Get structure (model 0) sequence"""
        
        trans = AATranslator()
        
        if chain_id in [None,True]:
            #all
            chains = self.get_prime_model().child_list
        else:
            chains = [self._get_chain_if_available(chain_id)]
        
        seq_list = []
        
        for cha in chains:
            seq=''
            for res in cha.get_iterator(SEQRES):
                resname = res.get_resname()
                if ORIGINAL:
                    aa = trans.get1original(res.get_resname())
                    if not UNKNOWN_ALLOWED and aa == trans.unknown1:
                        raise KeyError, res.get_resname()
                else:
                    aa = trans.get1letter(res.get_resname())
                seq=seq+aa
            seq_list.append(seq)
        
        if chain_id == True:
            return seq_list
        
        sequence = join(seq_list,ter_tag)
        
        return  sequence
    
    def get_fasta(self, name=None, chain_id=True, ter_tag='',ORIGINAL=0, SEQRES=0, UNKNOWN_ALLOWED=1):
        """Converts a structure's sequence into fasta obj"""
        
        gapchar = '-'
        
        if name == None:
            name = self.get_id()
        sequence = self.get_sequence(chain_id,ter_tag,ORIGINAL,SEQRES,UNKNOWN_ALLOWED)
        if chain_id != True:
            L = len(sequence)
            sequence = [sequence]
        else:
            L = 0
            for seq in sequence:
                if len(seq) > L:
                    L=len(seq)
        
        if ORIGINAL:
            alphabet=Gapped(IUPACProtein(), gapchar)
        else:
            alphabet=Gapped(ExtendedIUPACProtein(), gapchar)
        fasta = FastaAlign.FastaAlignment(alphabet)
        
        for i in range(len(sequence)):
            seq = sequence[i]
            if chain_id in [True,False]:
                cid = ':'+self.get_prime_model().child_list[i].get_id()
            elif chain_id == None:
                cid = ''
            else:
                cid = ':'+chain_id
            fasta.add_sequence(name+cid,seq+(L-len(seq))*gapchar)
        
        return fasta
        
    def save_sequence(self, filename=None,  name=None, chain_id=True, ter_tag='', ORIGINAL=0, SEQRES=0, UNKNOWN_ALLOWED=1):
        """Saves a structure's sequence into the file.
        
        File format is fasta.
        
        Options:
        o filename - if null, structure id is used
        o name     - for the sequence; if null, structure id is used
        o chain_id:  None - all chains together
                           True - chains as separate sequences
                           False - if chain is unique
                           char - specific chain
        o ter_tag  - in conjunction with chain_id=None
        
        Flags:
        ORIGINAL                     - converting modified resnames into original
        SEQRES                        - including seqres (missing coords) residua
        UNKNOWN_ALLOWED - if set to false in conjunction with ORIGINAL, exception is risen while unknown aa is found
        """
        
        brk=60
        fasta = self.get_fasta(name, chain_id, ter_tag, ORIGINAL, SEQRES, UNKNOWN_ALLOWED)
        
        if filename == None:
            if name == None:
                filename = self.get_id()+'.fas'
            else:
                filename = name+'.fas'
        
        f=open(filename,'w')
        for seq in fasta.get_all_seqs():
            f.write('>'+seq.description+'\n')
            sequence = seq.seq.tostring()
            i = 0
            while brk*i < len(sequence):
                f.write(sequence[i:(i+1)*brk]+'\n')
                i=i+1
        
        f.close()
        
    def save_with_remark(self, filename, text=None, header=None, score=None):
        """
        This method save the structure, enriching its header with REMARK 0 comments taken from given message string.
        
        Text string can have any length and forced eoln ('\n') characters, while all is formatted and fit into 80 character 
        PDB row length.
        
        If header (row list) is given, REMARK 0 is inserted into it as first REMARK records.
        
        Additionally, score (MQAP-typical) information can be given in following forms:
        -string (plain text)
        -value (integer or float)
        -method_name->value dictionary
        
        It is placed in a form of last added records - 'REMARK   0 SCORE:'...
        
        At the end, all PDB blocks are saved in following order: HEADER (REMARK within) + COORDINATES.
        """
        
        remark0 = []
        Max = 80
        
        start = 'REMARK    '
        if text != None:
            try:
                text_list = text.split(' ')
            except:
                raise Exception('text - string object expected not a  %s'%text)
            row = start
            i = 0
            j = 0
            while i < len(text_list):
                phrase=split(text_list[i],'\n')
                word = phrase[j]
                if phrase==['']:
                    j=0
                    i=i+1
                    continue
                j=j+1
                if j == len(phrase):
                    j=0
                    i=i+1
                    newline_word = False
                else:
                    newline_word = True
                length_w = len(word)
                length = len(row)
                if length + length_w + 1< Max:
                    #fits into line
                    row = row + ' ' + word
                else:
                    #exceeds line, eoln
                    row = row + (Max-length)*' '+'\n'
                    remark0.append(row)
                    row = start + ' '+ word
                length = len(row)
                if newline_word == True:
                    #eoln forced by a character
                    row = row + (Max-length)*' '+'\n'
                    remark0.append(row)
                    row = start
            length = len(row)
            row = row + (Max-length)*' '+'\n'
            remark0.append(row)
        
        #adding score information if available
        if score != None:
            if type(score) == StringType:
                #plaintext
                row = start+' SCORE: ' + score
                length = len(row)
                row = row + (Max-length)*' '+'\n'
                remark0.append(row)
            elif type(score) in [IntType, FloatType]:
                #numerical value
                row = start+' SCORE: ' + str(score)
                length = len(row)
                row = row + (Max-length)*' '+'\n'
                remark0.append(row)
            else:
                #dictionary - might be multiple
                methods = score.keys()
                methods.sort()
                for method in methods:
                    row = start+' SCORE- ' + method + ': '+ str(score[method])
                    length = len(row)
                    row = row + (Max-length)*' '+'\n'
                    remark0.append(row)
        
        if header != None:
            pointer=0
            while pointer < len(header) and header[pointer][:6] != 'REMARK':
                pointer=pointer+1
            header = header[:pointer]+remark0+header[pointer:]
        else:
            header = remark0
        
        io=PDBIO()
        io.set_structure(self)
        io.save(filename)
        structure_rowdata = open(filename,'r').readlines()
        rowdata = header + structure_rowdata
        rowdata = rowdata + ['END\n']
        f = open(filename,'w')
        f.writelines(rowdata)
        f.close()
        
        return True

    def substitute_bfactor(self, bfactor_dict_mixed, chain_id=None):
        """
        This method substitutes bfactor values for a given chain (model 0)
        
        Value dictionary can contain:
            -string (unparsed) resseq (residue indices)
            -resseq
            -residue_id tuple: (hetfield,resseq,icode)
        Unless they collide, all are converted into proper residue_ids.
        
        If multiple chains, particular chain is extracted (given chain_id) and all residues having its ids in converted value dictionary
        are routined by their atoms and all bfactor values are substituted.
        """
        
        bfactor_resid, cha = self._map_dict(bfactor_dict_mixed, chain_id)
        
        #proper substitution
        for res_key in bfactor_resid.keys():
            try:
                res=cha[res_key]
                bfactor = bfactor_resid[res_key]
                for a in res.get_iterator():
                    a.set_bfactor(bfactor)
            except KeyError, message:
                raise
            
        return True
    
    def substitute_occupancy(self, occupancy_dict_mixed, chain_id=None):
        
        """
        This method substitutes occupancy values for a given chain (model 0)
        
        Value dictionary can contain:
            -string (unparsed) resseq (residue indices)
            -resseq
            -residue_id tuple: (hetfield,resseq,icode)
        Unless they collide, all are converted into proper residue_ids.
        
        If multiple chains, particular chain is extracted (given chain_id) and all residues having its ids in converted value dictionary
        are routined by their atoms and all occupancy values are substituted.
        """
        
        occupancy_resid, cha = self._map_dict(occupancy_dict_mixed, chain_id)
        
        #proper substitution
        for res_key in occupancy_resid.keys():
            try:
                res=cha[res_key]
                occupancy = occupancy_resid[res_key]
                for a in res.get_iterator():
                    a.set_occupancy(occupancy)
            except KeyError, message:
                raise
            
        return True

##########################################################################

class Model(model):
    
    def add(self, c):
        try:
            return model.add(self,c)
        except PDBConstructionException, message:
            raise BPDBConstructionException, message
    
    def get_parent(self):
        """Returns the parent object
        
        A bugifx to biopython 1.41
        """
        if self.parent is None and self.get_level() != 'S':
            raise BPDBException, 'No parent'
        return self.parent
    
    def get_sequence(self, chain_id = None,ter_tag='',ORIGINAL=0, SEQRES=0,UNKNOWN_ALLOWED=1):
        """Get structure (model) sequence"""
        return self.get_parent().get_sequence(chain_id,ter_tag,ORIGINAL, SEQRES,UNKNOWN_ALLOWED)
        
##########################################################################

class Chain(chain):
    
    def __init__(self, id, POLYPEPTIDE=0):
        chain.__init__(self, id)
        
        if POLYPEPTIDE:
            self.xtra['polypeptide'] = True
            self.polypeptide = Polypeptide()
        else:
            self.polypeptide = None

    def __iter__(self):
        "Iterate over non-seqres children."
        return self.get_iterator(False)
    
    def _add_polypeptide(self, r):
        """Wraps new residue into polypeptide if it exists.
        
        Seqres entries are excluded.
        """
        if self.is_polypeptide():
            if r.is_disordered() != 3:
                self.polypeptide.append(r)
                return True
            return None
        return False
        
    def add(self, r):
        "Add a child. If polypeptide structure present, add there as well."
        try:
            chain.add(self,r)
            self._add_polypeptide(r)
        except PDBConstructionException, message:
            raise BPDBConstructionException, message
    
    def detach_child(self, id):
        "Remove a child. If polypeptide structure present, remove as well."
        try:
            if is_polypeptide():
                self.polypeptide.pop(self[id])
        except:
            #will be risen in the next step
            pass
        chain.detach_child(self, id)
        
    def get_iterator(self,SEQRES=0):
        "Iterate over children."
        for child in self.child_list:
            if SEQRES or not child.is_disordered() == 3:
                yield child    
    
    def get_iterator_seqres(self):
        "Iterate over any children."
        return self.get_iterator(True)
    
    def get_parent(self):
        """Returns the parent object
        
        A bugifx to biopython 1.41
        """
        if self.parent is None and self.get_level() != 'S':
            raise BPDBException, 'No parent'
        return self.parent
    
    def get_sequence(self,ORIGINAL=0, SEQRES=0,UNKNOWN_ALLOWED=1):
        """Returns the sequence of the chain."""
        chain_id = self.get_id()
        return self.get_parent().get_sequence(chain_id,'',ORIGINAL,SEQRES,UNKNOWN_ALLOWED)
    
    def is_polypeptide(self):
        """Boolean - returns if chain has polypeptide wrapper"""
        return self.polypeptide != None

##########################################################################

class Residue(residue):
    
    def __iter__(self):
        "Iterate over non-seqres children."
        return self.get_iterator(False)
    
    def add(self, a):
        try:
            return residue.add(self,a)
        except PDBConstructionException, message:
            raise BPDBConstructionException, message
        
    def get_iterator(self,SEQRES=0):
        "Iterate over children."
        for child in self.child_list:
            if SEQRES or not child.is_disordered() == 3:
                yield child    
    
    def get_iterator_seqres(self):
        "Iterate over any children."
        return self.get_iterator(True)
        
    def get_parent(self):
        """Returns the parent object
        
        A bugifx to biopython 1.41
        """
        if self.parent is None and self.get_level() != 'S':
            raise BPDBException, 'No parent'
        return self.parent
    
    def is_hetatm(self):
        "Boolean."
        return (self.get_id()[0] != ' ')

##########################################################################

class Atom(atom):

    def __init__(self, name, coord, bfactor, occupancy, altloc, fullname, serial_number=None, element=None):
        """Biopython - version independed scalable code."""
        try:
            #biopython 1.41
            atom.__init__(self, name, coord, bfactor, occupancy, altloc, fullname, serial_number, element)
        except TypeError:
            #old biopython
            atom.__init__(self, name, coord, bfactor, occupancy, altloc, fullname, serial_number)
        except TypeError:
            #old biopython
            atom.__init__(self, name, coord, bfactor, occupancy, altloc, fullname)

    def get_parent(self):
        """Returns the parent object
        
        A bugifx to biopython 1.41
        """
        if self.parent is None and self.get_level() != 'S':
            raise BPDBException, 'No parent'
        return self.parent
    
    def get_symbol(self):
        """Returns chemical symbol
        
        Based on a PDB spec.  Works fine for amino acids (non-hetams) and nucleic acids, as well
        """
        sym = self.get_fullname()[:2]
        
        try:
            i = int(sym[0])
            return sym[1]
        except ValueError:
            return strip(sym)
    
    def is_hydrogen(self):
        #Deprecated
        return (self.get_symbol() == 'H')
    
##########################################################################
##### new classes #############################################################
##########################################################################

class extendedStructure(Structure):
    #deprecated - renamed to Structure
    pass
    
##########################################################################
    
class MissingResidue(Residue):
    
    """This class is a Bio.PDB.Residue SMCRA object modification, allowing residue having no atoms (e.g. originating from SEQRES).
    
    It may also contain atoms with no coordinates.
    """
    
    def __init__(self, id, resname, segid, MAKECORE=1):
        
        self.core_atoms=['N','CA','C','O']
        
        Residue.__init__(self, id, resname, segid)
        if MAKECORE:
            self._core_fill()
        
        self.disordered=3
    
    def __iter__(self):
        "Iterate over non-seqres children."
        return self.get_iterator(False)
    
    def __repr__(self):
        resname=self.get_resname()
        hetflag, resseq, icode=self.get_id()
        full_id=(resname, hetflag, resseq, icode)
        return "<MissingResidue %s het=%s resseq=%s icode=%s>" % full_id
        
    def _core_fill(self):
        """Adds an empty core atoms (names only"""
        for name in self.core_atoms:
            self.add(CompletelyDisorderedAtom(name))
    
    def add(self, at):
        if isinstance(at,CompletelyDisorderedAtom):
            Residue.add(self,at)
            return self
        else:
            for a in self.get_iterator_seqres():
                if not a.is_disordered() == 3:
                    Residue.add(self,at)
                    return self
            #new Residue
            pure_residue = Residue(self.get_id(),self.get_resname(),self.get_segid())
            for a in self.get_iterator_seqres():
                if a.is_disordered() != 3:
                    self.detach_child(a)
                    pure_residue.add(a)
            pure_residue.add(at)
            parent = self.get_parent()
            self.get_parent().detach_child(self.get_id())
            parent.add(pure_residue)
            return pure_residue
    
    def get_iterator(self,SEQRES=0):
        "Iterate over children."
        for child in self.child_list:
            if SEQRES or not child.is_disordered() == 3:
                yield child    
    
    def get_iterator_seqres(self):
        "Iterate over seqres children."
        return self.get_iterator(True)
        
##########################################################################

class CompletelyDisorderedAtom(Atom):
    
    """This class is a Bio.PDB.Atom SMCRA object modification, allowing atom with no coordinates."""
    
    def __init__(self, name, occupancy=0.0, altloc=' ', fullname=None, serial_number=None, mode='swiss'):
            
        self.blank_coord_dictionary={'swiss': 9999.999,'rosetta':0.000}
        self.blank_bfactor_dictionary={'swiss': 99.99,'rosetta':0.00}
        
        coord = array((self.blank_coord_dictionary[mode],self.blank_coord_dictionary[mode],self.blank_coord_dictionary[mode],),'f')
        bfactor = self.blank_bfactor_dictionary[mode]
        if fullname==None:
            fullname=(2-len(name))*' '+name+'  '
        
        Atom.__init__(self, name, coord, bfactor, occupancy, altloc, fullname, serial_number)
        
        self.disordered_flag=3
        self.mode = mode
    
    def __repr__(self):
        "Print CompletelyDisorderedAtom object."
        return "<CompletelyDisorderedAtom %s in %s mode>" % (self.get_id(), self.get_mode())
    
    def __sub__(self, other):
        raise TypeError, "unsupported operand type(s) for -: 'instance' and '%s'"%(type(other).__name__)
    
    def get_mode(self):
        return self.mode
    
    def set_coord(self, coord):
        
        pure_atom = Atom(self.get_name(), coord, self.get_bfactor(), self.get_occupancy(), self.get_altloc(), self.get_fullname())
        parent = self.get_parent()
        self.get_parent().detach_child(self.get_id())
        parent.add(pure_atom)
        return pure_atom
    
    def switch_mode(self):
        modechoice = self.blank_coord_dictionary.keys()
        
        self.mode = modechoice[(modechoice.index(self.mode)+1) % len(modechoice)]
        self.coord = array((self.blank_coord_dictionary[self.mode],self.blank_coord_dictionary[self.mode],self.blank_coord_dictionary[self.mode],),'f')
        self.bfactor = self.blank_bfactor_dictionary[self.mode]
        
        return self.mode

##########################################################################
### extended io classes ###########################################################
##########################################################################

class PDBIO(pdbio):
    
    def __init__(self,use_model_flag=0,SINGLEMODEL=0,NOHETATM=0,SEQRES=0):
        """Flag parameters init.
        
        use_model_flag  - if 1, force use of the MODEL record in output.
        SINGLEMODEL   - only first model will be saved
        NOHETATM        - heteroatoms (residua having non-empty hetfield) will be rejected
        SEQRES             - SEQRES (having no coords) entities, either residua or atoms will be rejected.
                                    This is the case of MissingResidue and CompletelyDisorderedAtom classes.
        chain_id             - provided on structure fix; the only chain to be saved unless null.
        """
        self.SINGLEMODEL=SINGLEMODEL
        self.NOHETATM=NOHETATM
        self.SEQRES=SEQRES
        self.chain_id=None
        
        pdbio.__init__(self,use_model_flag)
    
    def set_structure(self, structure, chain_id = None):
        """Sets the structure to save and checks the chain_id eventually."""
        
        self.structure=structure
        if chain_id != None:
            try:
                c = self.structure.get_prime_model()[chain_id]
            except KeyError:
                #no such chain
                raise
            self.chain_id=chain_id
        
    def save(self, file=None, select=None):
        """Saves the structure into the file, using all predefined filters.
        
        If the filename is not given, it is derived form the
        """
        if select == None:
            select = Select(chain_id=self.chain_id,SINGLEMODEL=self.SINGLEMODEL,NOHETATM=self.NOHETATM,SEQRES=self.SEQRES)
        
        if file == None:
            try:
                structure_name = str(self.structure.get_id())
            except:
                #what tha ...?
                structure_name = 'structure'
            file =structure_name +'.pdb'
        
        return pdbio.save(self, file, select)
    
##########################################################################

class Select(select_io):
    
    def __init__(self,chain_id=None,SINGLEMODEL=0,NOHETATM=0,SEQRES=0):
        """Flag parameters init."""
        
        self.SINGLEMODEL=SINGLEMODEL
        self.NOHETATM=NOHETATM
        self.SEQRES=SEQRES
        self.chain_id = None
    
    def __repr__(self):
        state = ''
        
        if self.chain_id != None:
            state=state+'chain '+self.chain_id+' '
        if self.SINGLEMODEL:
            state=state+'SINGLEMODEL '
        if self.NOHETATM:
            state=state+'NOHETATM '
        if self.SEQRES:
            state=state+'SEQRES '
        
        if len(state) == 0:
            #none
            state = 'all'
        return "<Select %s>"%(state)
        
    def accept_model(self, mdl):
        """Filters single model only."""
        
        if not self.SINGLEMODEL:
            return 1
        elif model is mdl.get_parent().get_prime_model():
            return 1
        else:
            return 0
    
    def accept_chain(self, cha):
        """Filters by specific chain id."""
        
        if self.chain_id == None:
            return 1
        elif cha.get_id() == self.chain_id:
            return 1
        else:
            return 0
        
    def accept_residue(self, res):
        """Filters aminoacid and non-missing residua."""
        
        if self.NOHETATM and res.get_id()[0] != ' ':
            return 0
        if self.SEQRES and res.is_disordered() == 3:
            return 0
        return 1
    
    def accept_atom(self, at):
        """Filters non- completely disordered atoms."""
        
        if self.SEQRES and at.is_disordered() == 3:
            return 0
        return 1

##########################################################################
### aux & conversion functions
##########################################################################

def clean_coords(node,disorder=1):
    """This function cleans the given SMCRA entity from seqres (no coords) elements (MissingResidues or CompletelyDisorderedAtoms)."""
    if node.get_level() == 'A':
        return node
    elif node.get_level() in ['R','C']:
        for child in node.get_list():
            if child.is_disordered() >= disorder:
                #no coords - remove
                node.detach_child(child)
        return node
    else:
        for child in node.get_iterator():
            clean_coords(child)
    
    return node

##########################################################################

def extend_structure(struct, id=None):
    """Converts Bio.PDB.Structure instance to extendedStructure object"""
    
    if isinstance(struct,structure) or isinstance(struct,Structure):
        if id == None:
            id = struct.get_id()
        #estructure=extendedStructure(id)
        estructure=Structure(id)
        
        for mdl in struct.child_list:
            mdl.detach_parent()#unnecessary
            struct.detach_child(mdl.get_id())
            estructure.add(mdl)
            mdl.set_parent(estructure)
    
        return estructure
    
    raise TypeError, 'extend_structure struct argument must be a Bio.PDB.Structure.Structure or genesilico.structure.SMCRA_extended.Structure instance'

##########################################################################

def make_polypeptide(entity, radius=0, METHOD='ca', SMCRA=0):
    """Given SMCRA entity (S M or C) the function copies and returns polypeptide - based structure.
    
    radius - required only in 'any' method
    o METHOD:  'ca' (CA -based)
                        'cn' (C-N -based)
                        'any' (any atoms).
    o SMCRA: 0 - Bio.PDB.Polypeptide.Polypeptide object is returned
                    1 - Structure object is returned
                    2 - same as (1) but with polypeptide chain wrapper
    """
    if METHOD == 'ca':
        if radius == 0:
            R = 4.3
        else:
            R = radius
        builder = CaPPBuilder(radius=R)
    elif METHOD == 'cn':
        if radius == 0:
            R = 1.8
        else:
            R = radius
        builder = PPBuilder(radius=R)
    else:
        #any
        builder = _PPBuilder(radius=radius)
    
    if entity.get_id() != 'S':
        new_entity = clean_coords(deepcopy(entity))
    else:
        #to speed up
        base_entity = Structure(entity.get_id())
        new_model = clean_coords(deepcopy(entity.get_prime_model()))
        base_entity.add(new_model)
        
    try:
        peptides = builder.build_peptides(new_entity)
    except PDBException, message:
        raise BPDBException, message
    
    if not SMCRA:
        return peptides
    
    s_id = ''
    for eid in entity.get_full_id():
        if s_id != '':
            s_id = s_id+':'
        try:
            s_id = s_id+str(eid)
        except:
            pass
    
    stru = Structure(s_id)
    mdl = Model(0)
    stru.add(mdl)
    for i in range(len(peptides)):
        chain_id = ascii_uppercase[i]
        cha = Chain(chain_id)
        mdl.add(cha)
        for res in peptides[i]:
            cha.add(res)
        if SMCRA == 2:
            cha.xtra['polypeptide'] = True
            cha.polypeptide = peptides[i]
        
    return stru
